
#include <unistd.h>
#include <fcntl.h>
#include "comm.h"
#include "../constable.h"
#include "../medusa_object.h"
#include "../event.h"
//#include "../language/language.h"
#include <sys/param.h>

#include <sys/socket.h>
#include <netinet/in.h>

//#define DEBUG_COMM

static char comm_buf[8192];
static char comm_obuf[1024];

int read_all( int fd, void *buf, int count )
{ int l,r;
	l=0;
	while( count>0 )
	{	r=read(fd,buf,count);
		if( r<=0 )
			return(r);
		count-=r;
		buf+=r;
		l+=r;
	}
	return(l);
}

#ifdef DEBUG_COMM
#define	Read	read_print
#include <ctype.h>
int read_print( int fd, void *buf, int count )
{ int r;
  int i;
	printf("read(%d,%p,%d)",fd,buf,count);
	r=read_all(fd,buf,count);
	printf("=%d \"",r);
	for(i=0;i<r;i++)
		if( ((char*)buf)[i] == '\\' )
			printf("\\");
		else if( isprint(((char*)buf)[i]) )
			printf("%c",((char*)buf)[i]);
		else	printf("\\x%02d",((char*)buf)[i]);
	printf("\"\n");
	return(r);
}
#else
#define	Read	read_all
#endif

static int get_event_context( struct event_context_s *c, struct event_type_s *t, void *data )
{
	c->operation.next=&(c->subject);
	c->operation.attr.offset=0;
	c->operation.attr.length=t->m.size;
	c->operation.attr.type=MED_TYPE_END;
	strncpy(c->operation.attr.name,t->m.name,MIN(MEDUSA_ATTRNAME_MAX,MEDUSA_OPNAME_MAX));
	c->operation.flags=0;
	c->operation.class=&(t->operation_class);
	c->operation.data=(char*)(data)+2*sizeof(u_int32_t);

	c->subject.next=&(c->object);
	c->subject.attr.offset=c->subject.attr.length=0;
	c->subject.attr.type=MED_TYPE_END;
	strncpy(c->subject.attr.name,t->m.op_name[0],MEDUSA_ATTRNAME_MAX);
	c->subject.flags=0;
	c->subject.class=t->op[0];
	c->subject.data=(char*)(data)+2*sizeof(u_int32_t)+(t->m.size);
	c->object.next=NULL;
	c->object.attr.offset=c->object.attr.length=0;
	c->object.attr.type=MED_TYPE_END;
	strncpy(c->object.attr.name,t->m.op_name[1],MEDUSA_ATTRNAME_MAX);
	c->object.flags=0;
	c->object.class=t->op[1];
	c->object.data=(char*)(data)+2*sizeof(u_int32_t)+(t->m.size);
	if( c->subject.class!=NULL )
	{	c->object.data+= c->subject.class->m.size;
		c->subject.attr.length=c->subject.class->m.size;
	}
	if( c->object.class!=NULL )
	{
		c->object.attr.length=c->object.class->m.size;
	}
	c->local_vars=NULL;
	return(0);
}

static int comm_fd=-1;

int comm_open( char *filename )
{
	if( filename[0]=='#' )
	{ struct sockaddr_in addr;
	  int sock,len;
		if( (sock=socket(PF_INET,SOCK_STREAM,0))<0 )
			return(init_error("Can't create socket"));
		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr=INADDR_ANY;
		addr.sin_port=htons(atoi(filename+1));
		if( bind(sock,(struct sockaddr*)&addr,sizeof(struct sockaddr_in))<0 )
		{	close(sock);
			return(init_error("Can't bind to port %d",atoi(filename+1)));
		}
		listen(sock,1);
		len=sizeof(struct sockaddr_in);
		if( (comm_fd=accept(sock,(struct sockaddr *)&addr,&len))<0 )
		{	close(sock);
			return(init_error("Can't accept on port %d",atoi(filename+1)));
		}
		close(sock);
	}
	else if( (comm_fd=open(filename,O_RDWR))<0 )
	{	init_error("Can't open %s",filename);
		return(-1);
	}

	return(0);
}

static void *read_from_kernel( int len )
{ char *buf;
	if( len+2*sizeof(u_int32_t) > sizeof(comm_buf) )
	{	if( (buf=malloc(len+2*sizeof(u_int32_t)))==NULL )
		{	fatal(Out_of_memory);
			return(NULL);
		}
		memcpy(buf,comm_buf,2*sizeof(u_int32_t));
	}
	else	buf=comm_buf;
	if( Read(comm_fd,buf+2*sizeof(u_int32_t),len) != len )
	{	fatal("Data read error!");
		close(comm_fd);
		return(NULL);
	}
	return((void*)buf);
}

static void *cont_read_from_kernel( void *buf, int ofs, int len )
{ char *b2;
	if( buf==comm_buf && ofs+len+2*sizeof(u_int32_t) > sizeof(comm_buf) )
	{	if( (buf=malloc(ofs+len+2*sizeof(u_int32_t)))==NULL )
		{	fatal(Out_of_memory);
			return(NULL);
		}
		memcpy(buf,comm_buf,2*sizeof(u_int32_t));
	}
	else if( buf!=comm_buf )
	{	if( (buf=realloc((b2=buf),ofs+len+2*sizeof(u_int32_t)))==NULL )
		{	free(b2);
			fatal(Out_of_memory);
			return(NULL);
		}
	}
	if( Read(comm_fd,buf+ofs+2*sizeof(u_int32_t),len) != len )
	{	fatal("Data read error!");
		close(comm_fd);
		return(NULL);
	}
	return((void*)buf);
}

static void *read_attr_from_kernel( int len )
{ void *buf;
	if( (buf=read_from_kernel(len))==NULL )	/* malloc */
		return(NULL);
	while( ((struct medusa_comm_attribute_s *)(buf+2*sizeof(u_int32_t)+len-sizeof(struct medusa_comm_attribute_s)))->type!=MED_TYPE_END )
	{	if( (buf=cont_read_from_kernel(buf,len,sizeof(struct medusa_comm_attribute_s)))==NULL )
			return(NULL);
		len+=sizeof(struct medusa_comm_attribute_s);
	}
	return(buf);
}

static void free_from_kernel( void *buf )
{
	if( buf!=comm_buf )
		free(buf);
}

int comm_do( int(*fn)(void *), void *arg )
{ struct event_type_s *t;
  struct event_context_s c;
  char *buf;
  int result;
  int len;
  struct class_s *cl;
//printf("ZZZ: snazim sa nieco nacitat\n");
	if( Read(comm_fd,comm_buf,2*sizeof(u_int32_t)) != 2*sizeof(u_int32_t) )
	{	fatal("Read error!");
		close(comm_fd);
		return(-1);
	}
	if( ((u_int32_t*)comm_buf)[0]!=0 )
  	{	
//printf("ZZZ: mam poziadavku\n");
//printf("ZZZ: idem riesit poziadavku\n");
		if( (t=event_type_find( ((u_int32_t*)comm_buf)[0]))==NULL )
		{	fatal("comm: Unknown access type 0x%08x!",((u_int32_t*)comm_buf)[0]);
			return(-1);
		}
	        len= (t->m.size);
	        if( t->op[0]!=NULL )
	                len += t->op[0]->m.size;
	        if( t->op[1]!=NULL )
	                len += t->op[1]->m.size;
//printf("ZZZ: docitavam poziadavku\n");
		if( (buf=read_from_kernel(len))==NULL )	/* malloc */
			return(-1);
		if( fn!=NULL )
		{	if( fn(arg)<0 )
				return(-1);
		}
//printf("ZZZ: riesim poziadavku\n");
		get_event_context( &c, t, buf );
		result=do_event( &c, t );
//printf("ZZZ: vyriesene\n");
		/* TODO: only if changed */
		if( result >=0 )
			comm_update_object( &(c.subject) );
		free_from_kernel( buf );
		((u_int32_t*)(comm_obuf))[0]= MEDUSA_COMM_AUTHANSWER;
		((u_int32_t*)(comm_obuf))[1]= ((u_int32_t*)(comm_buf))[1];
#ifdef TRANSLATE_RESULT
		switch( result )
		{	case RESULT_ALLOW:
				*((u_int16_t*)(comm_obuf+2*sizeof(u_int32_t)))
					= MED_YES;
				break;
			case RESULT_DENY:
				*((u_int16_t*)(comm_obuf+2*sizeof(u_int32_t)))
					= MED_NO;
				break;
			case RESULT_SKIP
				*((u_int16_t*)(comm_obuf+2*sizeof(u_int32_t)))
					= MED_SKIP;
				break;
			case RESULT_OK:
				*((u_int16_t*)(comm_obuf+2*sizeof(u_int32_t)))
					= MED_OK;
				break;
			default:
				*((u_int16_t*)(comm_obuf+2*sizeof(u_int32_t)))
					= MED_ERR;
		}
#else
		*((u_int16_t*)(comm_obuf+2*sizeof(u_int32_t))) = result;
#endif
		len=2*sizeof(u_int32_t)+sizeof(u_int16_t);
//printf("ZZZ: odpovedam\n");
		if( write(comm_fd,comm_obuf,len)!=len )
		{	fatal("Write error!");
			close(comm_fd);
			return(-1);
		}
		return(1);
	}
	else switch( ((u_int32_t*)comm_buf)[1] )
	{
	case MEDUSA_COMM_CLASSDEF:
printf("ZZZ: MEDUSA_COMM_CLASSDEF\n");
		len=sizeof(struct medusa_comm_class_s)+sizeof(struct medusa_comm_attribute_s);
		if( (buf=read_attr_from_kernel(len))==NULL )	/* malloc */
			return(-1);
		if( (cl=add_class((struct medusa_class_s*)(buf+2*sizeof(u_int32_t)),
			(struct medusa_attribute_s*)(buf+2*sizeof(u_int32_t)+sizeof(struct medusa_comm_class_s))))==NULL )
			runtime("Can't add class");
		free_from_kernel(buf);
		break;
	case MEDUSA_COMM_ACCTYPEDEF:
printf("ZZZ: MEDUSA_COMM_ACCTYPEDEF\n");
		len=sizeof(struct medusa_comm_acctype_s)+sizeof(struct medusa_comm_attribute_s);
		if( (buf=read_attr_from_kernel(len))==NULL )	/* malloc */
			return(-1);
		if( event_type_add((struct medusa_acctype_s*)(buf+2*sizeof(u_int32_t)),
			(struct medusa_attribute_s*)(buf+2*sizeof(u_int32_t)+sizeof(struct medusa_comm_acctype_s)))<0 )
			runtime("Can't add acctype");
		free_from_kernel(buf);
		break;
	case MEDUSA_COMM_CLASSUNDEF:
	case MEDUSA_COMM_ACCTYPEUNDEF:
printf("ZZZ: UNDEF\n");
		runtime("Huh, I don't know how to undef class or acctype ;-|");
		if( (buf=read_from_kernel(sizeof(u_int32_t)))==NULL )	/* malloc */
			return(-1);
		free_from_kernel(buf);
		break;
	default:
		fatal("Communication protocol error! (%d)",((u_int32_t*)comm_buf)[1]);
		break;
	}
	return(0);
}

int comm_fetch_object( struct object_s *o )
{ static u_int32_t id=2;
  u_int32_t ans[4];
	/* POZOR!!! moze byt zavolane len medzi otazkou a odpovedou. */
	((u_int32_t*)(comm_obuf))[0]= MEDUSA_COMM_FETCH_REQUEST;
	((u_int32_t*)(comm_obuf))[1]= o->class->m.classid;
	((u_int32_t*)(comm_obuf))[2]= id++;
printf("ZZZ: pytam objekt\n");
	if( write(comm_fd,comm_obuf,3*sizeof(u_int32_t))!=3*sizeof(u_int32_t) )
	{	fatal("Write error!");
		close(comm_fd);
		return(-1);
	}
//printf("ZZZ: stale pytam objekt\n");
	if( write(comm_fd,o->data,o->class->m.size)!=o->class->m.size )
	{	fatal("Write error!");
		close(comm_fd);
		return(-1);
	}
//printf("ZZZ: citam objekt\n");
	if( Read(comm_fd,ans,4*sizeof(u_int32_t)) != 4*sizeof(u_int32_t) )
	{	fatal("Read error!");
		close(comm_fd);
		return(-1);
	}
	if( ans[1]==MEDUSA_COMM_FETCH_ERROR )
	{	runtime("fetch failed");
		return(-1);
	}
	if( ans[0]!=0 || ans[1]!=MEDUSA_COMM_FETCH_ANSWER || ans[2]!=o->class->m.classid
		|| ans[3]!= id-1 )
	{	fatal("Bad answer!");
		close(comm_fd);
		return(-1);
	}
//printf("ZZZ: stale citam objekt\n");
	if( Read(comm_fd,o->data,o->class->m.size)!=o->class->m.size )
	{	fatal("Fetch object error!");
		close(comm_fd);
		return(-1);
	}
//printf("ZZZ: nacitany objekt\n");
	return(0);
}

int comm_update_object( struct object_s *o )
{ 
	((u_int32_t*)(comm_obuf))[0]= MEDUSA_COMM_UPDATE;
	((u_int32_t*)(comm_obuf))[1]= o->class->m.classid;
printf("ZZZ: posielam objekt\n");
	if( write(comm_fd,comm_obuf,2*sizeof(u_int32_t))!=2*sizeof(u_int32_t) )
	{	fatal("Write error!");
		close(comm_fd);
		return(-1);
	}
//printf("ZZZ: stale posielam objekt\n");
	if( write(comm_fd,o->data,o->class->m.size)!=o->class->m.size )
	{	fatal("Write error!");
		close(comm_fd);
		return(-1);
	}
//printf("ZZZ: poslal som objekt\n");
	return(0);
}


int comm_vs_size( void )
{ struct class_s *p;
  struct medusa_attribute_s *a;
	if( (p=get_class_by_name("process"))==NULL )
	{	fatal("Undefined class 'process'!");
		return(-1);
	}
	if( (a=get_attribute(p,"vs"))==NULL )
	{	fatal("Class 'process' don't have attribute 'vs'!");
		return(-1);
	}
	return( a->length * 8 );
}

