
#include "constable.h"
#include "event.h"
#include "tree.h"
#include <sys/param.h>

int event_mask_clear( event_mask_t *e )
{
	memset(e->bitmap,0,sizeof(e->bitmap));
	return(0);
}

int event_mask_clear2( event_mask_t *e )
{
	memset(e->bitmap,0,2*sizeof(e->bitmap));
	return(0);
}

int event_mask_setbit( event_mask_t *e, int b )
{
	if( b<0 || b>=sizeof(e->bitmap)*8 )
		return(-1);
	setbit(e->bitmap,b);
	return(0);
}

int event_mask_clrbit( event_mask_t *e, int b )
{
	if( b<0 || b>=sizeof(e->bitmap)*8 )
		return(-1);
	clrbit(e->bitmap,b);
	return(0);
}

int event_mask_copy( event_mask_t *e, event_mask_t *f )
{ 
	memcpy(e->bitmap,f->bitmap,sizeof(e->bitmap));
	return(0);
}

int event_mask_copy2( event_mask_t *e, event_mask_t *f )
{
	memcpy(e->bitmap,f->bitmap,2*sizeof(e->bitmap));
	return(0);
}

int event_mask_and( event_mask_t *e, event_mask_t *f )
{ int i;
	for(i=sizeof(e->bitmap)-1;i>=0;i--)
		e->bitmap[i] &= f->bitmap[i];
	return(0);
}

int event_mask_or( event_mask_t *e, event_mask_t *f )
{ int i;
	for(i=sizeof(e->bitmap)-1;i>=0;i--)
		e->bitmap[i] |= f->bitmap[i];
	return(0);
}

int event_mask_or2( event_mask_t *e, event_mask_t *f )
{ int i;
	for(i=2*sizeof(e->bitmap)-1;i>=0;i--)
		e->bitmap[i] |= f->bitmap[i];
	return(0);
}

int event_mask_sub( event_mask_t *e, event_mask_t *f )
{ int i;
	for(i=sizeof(e->bitmap)-1;i>=0;i--)
		e->bitmap[i] &= ~(f->bitmap[i]);
	return(0);
}

struct event_type_s *global_event_types=NULL;
static struct event_type_s *last_event_type=NULL;

static struct event_type_s **event_type_hash=NULL;
static int event_type_hash_len,event_type_hash_ofs,event_type_hash_k;

struct event_type_s *event_from_medusa( struct medusa_acctype_s *m, struct medusa_attribute_s *a )
{ struct event_type_s *p;
  int l;
printf("ZZZ event_from_medusa:\n"); attr_print(a);
printf("ZZZ 0x%04x\n",m->actbit);
	for(l=0;a[l].type!=MED_TYPE_END;l++);
	l++;
	l*=sizeof(struct medusa_attribute_s);
	if( (p=malloc(sizeof(struct event_type_s)+l))==NULL )
		return(NULL);
	memcpy(&(p->m),m,sizeof(struct medusa_acctype_s));
	memset(&(p->operation_class),0,sizeof(p->operation_class));
//	p->operation_class.next=NULL;
//	p->operation_class.cinfo_offset=p->operation_class.cinfo_size=0;
//	p->operation_class.cinfo_mask=0;
//	p->operation_class.set=NULL;
//	p->operation_class.event_offset=p->operation_class.event_size=0;
//	p->operation_class.m.classid=0;
	p->operation_class.m.size=p->m.size;
	p->operation_class.m.name[0]=0;
	strncpy(p->operation_class.m.name,p->m.name,MIN(MEDUSA_CLASSNAME_MAX,MEDUSA_OPNAME_MAX));
	memcpy(p->operation_class.attr,a,l);
class_print(&(p->operation_class));
	if( (p->op[0]=get_class_by_id(p->m.op_class[0]))==NULL )
		init_error("Subject of access %s does not exist!",p->m.name);
	if( p->m.op_class[0]==p->m.op_class[1] && !strncmp(p->m.op_name[0],p->m.op_name[1],MEDUSA_ATTRNAME_MAX) )
		p->op[1]=NULL;
	else	p->op[1]=get_class_by_id(p->m.op_class[1]);
	event_mask_clear2(&(p->mask[0]));
	if( p->m.actbit&0x8000 )
		p->object=p->op[1];	/* object */
	else
		p->object=p->op[0];	/* subject */
	if( p->m.actbit&0x4000 )
		event_mask_setbit(&(p->mask[1]),(p->m.actbit)&0x00ff);
	else
		event_mask_setbit(&(p->mask[0]),(p->m.actbit)&0x00ff);
	p->handlers_hash=NULL;
	p->next=NULL;
	return(p);
}

int event_type_add( struct medusa_acctype_s *m, struct medusa_attribute_s *a )
{ struct event_type_s *p;
	if( (p=event_from_medusa(m,a))==NULL )
		return(-1);
	if( last_event_type==NULL )
		global_event_types=p;
	else	last_event_type->next=p;
	last_event_type=p;
	if( event_type_hash!=NULL )
		free(event_type_hash);
	event_type_hash=NULL;
	return(0);
}

int event_type_hash_create( void )
{ struct event_type_s *p;
  u_int32_t max,i;
	if( event_type_hash!=NULL )
		free(event_type_hash);
	max=0;
	event_type_hash_len=1;
	event_type_hash_ofs= (u_int32_t)(-1);
	for(p=global_event_types;p!=NULL;p=p->next)
	{	event_type_hash_len++;
		if( p->m.opid < event_type_hash_ofs )
			event_type_hash_ofs=p->m.opid;
		if( p->m.opid > max )
			max=p->m.opid;
	}
	event_type_hash_len*=4;
	event_type_hash_k= event_type_hash_len/(max-event_type_hash_ofs);

	if( (event_type_hash=malloc(event_type_hash_len*sizeof(struct event_type_s *)))==NULL )
		return(-1);
	memset(event_type_hash,0,event_type_hash_len*sizeof(struct event_type_s *));
	for(p=global_event_types;p!=NULL;p=p->next)
//		if( p->alt==NULL )
		{	i=(p->m.opid-event_type_hash_ofs)*event_type_hash_k;
			while( event_type_hash[i]!=NULL )
			{	i++;
				if( i>=event_type_hash_len )
					i=0;
			}
			event_type_hash[i]=p;
		}
	return(0);
}

struct event_type_s *event_type_find( u_int32_t opid )
{ int x;
	if( event_type_hash==NULL )
		event_type_hash_create();
	if( (x=(opid-event_type_hash_ofs)*event_type_hash_k)<event_type_hash_len )
	{	while( event_type_hash[x]!=NULL )
		{	if( event_type_hash[x]->m.opid==opid )
				return(event_type_hash[x]);
			x++;
			if( x>=event_type_hash_len )
				x=0;
		}
	}
	return(NULL);
}

struct event_type_s *event_type_find_name( char *name )
{ struct event_type_s *t;
	for(t=global_event_types;t!=NULL;t=t->next)
		if( !strncmp(t->m.name,name,MEDUSA_OPNAME_MAX) )
			break;
	return(t);
}

struct event_hadler_hash_s *evhash_add( struct event_hadler_hash_s **hash, struct event_handler_s *handler )
{ struct event_hadler_hash_s *l;
	if( (l=malloc(sizeof(struct event_hadler_hash_s)))==NULL )
	{	errstr=Out_of_memory;
		return(NULL);
	}
	for(;*hash!=NULL;hash=&((*hash)->next))
	{	if( (*hash)->handler==handler )
		{	free(l);
			return(*hash);
		}
	}
	l->next=NULL;
	l->handler=handler;
	*hash=l;
	return(l);
}


int register_event_handler( struct event_handler_s *h, struct event_type_s *type, vs_t *subject_vs, vs_t *object_vs )
{ struct event_hadler_hash_s *l;
  struct event_type_s *ftype=type;
//  	while( ftype->alt )	
//		ftype=ftype->alt;
	if( (l=evhash_add(&(ftype->handlers_hash),h))==NULL )
		return(-1);
	l->type=type;
	if( subject_vs==ALL_OBJ || (subject_vs==NULL && ftype->op[0]!=NULL) )
		vs_fill(l->subject_vs);
	else if( subject_vs )
		vs_set(subject_vs,l->subject_vs);
	else	vs_clear(l->subject_vs);
	if( object_vs==ALL_OBJ || (object_vs==NULL && ftype->op[1]!=NULL) )
		vs_fill(l->object_vs);
	else if( object_vs )
		vs_set(object_vs,l->object_vs);
	else	vs_clear(l->object_vs);
//printf("Zaregistrovane %p [%s]\n",l,l->type->m.name);
	return(0);
}


int evaluate_result( int old, int new )
{
/*
	if( new==RESULT_ALLOW && (old==RESULT_UNKNOWN || old==RESULT_OK) )
		return(new);
	else if( new==RESULT_DENY )
		return(new);
	else if( new==RESULT_SKIP && old!=RESULT_DENY )
		return(new);
	else if( new==RESULT_OK && old==RESULT_UNKNOWN )
		return(new);
*/
	if( old==RESULT_DENY || new==RESULT_DENY )
		return(RESULT_DENY);
	if( old==RESULT_UNKNOWN || new==RESULT_SKIP )
		return(new);
	if( old==RESULT_OK && new==RESULT_ALLOW )
		return(new);
	return(old);
}

int do_event( struct event_context_s *c, struct event_type_s *t )
{ struct event_handler_s *h;
  struct event_hadler_hash_s *hh;
  struct event_context_s ct;
  vs_t *v_s,*v_o;
  
//printf("ZZZZ1\n");
//  	while( t->alt )
//		t=t->alt;
//printf("ZZZZ2\n");

if( t->m.name[0]!=0 /* nie iget na subory */ )
{
printf("ZZZ: event: op=%s ",t->m.name);
if( c->subject.class!=NULL )
  printf("subject=%s ",c->subject.class->m.name);
else
  printf("subject=NONE ");
if( c->object.class!=NULL )
  printf("object=%s ",c->object.class->m.name);
else
  printf("object=NONE ");
printf("\n");
}
 
	c->result=RESULT_UNKNOWN;
	c->first=1;
//	c->local_vars=NULL;	/* ???? */
//printf("ZZZ: do_event 1\n");
	if( object_is_invalid(&(c->subject)) )
		object_do_sethandler(&(c->subject));	/* !!! ??? ZZZ */
//printf("ZZZ: do_event 2\n");
	if( c->subject.class!=NULL && c->subject.class->vs_size>0 )
		v_s=(vs_t*)(c->subject.data+c->subject.class->vs_offset);
	else
		v_s=NULL;
	if( c->object.class!=NULL && c->object.class->vs_size>0 )
		v_o=(vs_t*)(c->object.data+c->object.class->vs_offset);
	else
		v_o=NULL;

//printf("ZZZ: do_event 3\n");
	evhash_foreach(hh,t->handlers_hash)
	{	h=hh->handler;
//printf("testujem handler %p v_s=%p v_o=%p ic_s=%d ic_o=%d\n",hh,v_s,v_o,
//vs_isclear(hh->subject_vs),vs_isclear(hh->object_vs));
		if( (v_s==NULL) != (vs_isclear(hh->subject_vs)!=0) )
			continue;
		if( (v_o==NULL) != (vs_isclear(hh->object_vs)!=0) )
			continue;
//printf("hh->subject_vs=0x%08x ? v_s=0x%08x\n",hh->subject_vs[0],v_s[0]);
		if( v_s!=NULL && !vs_isfull(hh->subject_vs) )
		{	if( !vs_test(hh->subject_vs,v_s) )
				continue;
		}
//printf("hh->object_vs=0x%08x ? v_o=0x%08x\n",hh->object_vs[0],v_o[0]);
		if( v_o!=NULL && !vs_isfull(hh->object_vs) )
		{	if( !vs_test(hh->object_vs,v_o) )
				continue;
		}
//printf("handler\n");
		ct=*c;
		ct.result=RESULT_UNKNOWN;
//		ct.local_vars=h->local_vars;
		h->handler(h,&ct);
		c->result=evaluate_result(c->result,ct.result);
		c->first=0;
	}
//if( c.first ) printf("do_event: No event executed! [%s]\n",t->m.name);
//printf("ZZZ: do_event return %d\n",c->result);
if( function_debug!=NULL )
{
ct=*c;
strcpy(ct.operation.attr.name,"operation");
strcpy(ct.subject.attr.name,"subject");
strcpy(ct.object.attr.name,"object");
function_debug->handler(function_debug,&ct);
}
	return(c->result);
}

int get_empty_context( struct event_context_s *c )
{
	c->operation.next=&(c->subject);
	c->operation.attr.offset=c->operation.attr.length=0;
	c->operation.attr.type=MED_TYPE_END;
	c->operation.attr.name[0]=0;
	c->operation.flags=0;
	c->operation.class= NULL;
	c->operation.data=NULL;

	c->subject.next=&(c->object);
	c->subject.attr.offset=c->subject.attr.length=0;
	c->subject.attr.type=MED_TYPE_END;
	c->subject.attr.name[0]=0;
	c->subject.flags=0;
	c->subject.class= NULL;
	c->subject.data=NULL;
	c->object.next=NULL;
	c->object.attr.offset=c->object.attr.length=0;
	c->object.attr.type=MED_TYPE_END;
	c->object.attr.name[0]=0;
	c->object.flags=0;
	c->object.class= NULL;
	c->object.data=NULL;
	c->local_vars=NULL;
	return(0);
}

