
#include "execute.h"
#include "language.h"
#include "../constable.h"
#include "../tree.h"
#include "../space.h"
#include "../generic.h"
#include <unistd.h>

static void cmd_constable_pid( struct event_context_s *c, struct register_s *r, int(*getarg)(struct register_s*) )
{
	*((u_int32_t*)(r->data))=(u_int32_t)getpid();
}

static void cmd_hex( struct event_context_s *c, struct register_s *ret, int(*getarg)(struct register_s*) )
{ long long d;
  struct register_s r;
	ret->attr=&(execute_attr_str);
	ret->data[0]=0;
	if( !getarg(&r) || ((r.attr->type&0x0f)!=MED_TYPE_SIGNED && (r.attr->type&0x0f)!=MED_TYPE_UNSIGNED) )
	{	runtime("hex(): 1st argument must be signed or unsigned int");
		return;
	}
	object_get_val(r2o(&r),r.attr,&d,sizeof(d));
	sprintf(ret->data,"%llx",d);
	if( getarg(&r) )
	{	runtime("hex(): too many arguments");
		return;
	}
	return;
}

/* primaryspace(process,@"strom") */
static void cmd_primaryspace( struct event_context_s *c, struct register_s *ret, int(*getarg)(struct register_s*) )
{ 
  struct register_s r;
  struct object_s *o;
  struct tree_s *t;
  struct space_s *space;
	
	ret->attr=&(execute_attr_str);
	ret->data[0]=0;

	if( !getarg(&r) || (o=r.object)==NULL )
	{	runtime("primaryspace: 1st argument must be kobject");
		return;
	}
	if( !getarg(&r) || r.attr!=&execute_attr_pointer )
	{	runtime("primaryspace: 2nd argument must be path");
		return;
	}
	t=(struct tree_s*)(*((u_int32_t*)(r.data)));
	if( getarg(&r) )
	{	runtime("primaryspace: too many arguments");
		return;
	}
	if( t->type->obj_class != o->class )
	{	runtime("primaryspace: !!!!!! class mismatch");
		return;
	}
	t=(struct tree_s*)(*((u_int32_t*)(o->data+t->type->cinfo_offset)));
	if( (space=t->primary_space)==NULL )
		space=t->recursive_primary_space;
	if( space!=NULL )
		strncpy(ret->data,space->name,MAX_REG_SIZE);
	return;
}

/* enter(process,@"kam") */
static void cmd_enter( struct event_context_s *c, struct register_s *ret, int(*getarg)(struct register_s*) )
{ 
  struct register_s r;
  struct object_s *o;
  struct tree_s *t;
  int perm;
  vs_t vse[MAX_VS_BITS/32];
	
	if( !getarg(&r) || (o=r.object)==NULL )
	{	runtime("enter: 1st argument must be kobject");
		return;
	}
	if( !getarg(&r) || r.attr!=&execute_attr_pointer )
	{	runtime("enter: 2nd argument must be path");
		return;
	}
	t=(struct tree_s*)(*((u_int32_t*)(r.data)));
	if( getarg(&r) )
	{	runtime("primaryspace: too many arguments");
		return;
	}
	if( t->type->obj_class != o->class )
	{	runtime("primaryspace: !!!!!! class mismatch");
		return;
	}

	perm=1;
	if( (t->type->flags & GFL_USE_VSE) )
	{	object_get_vs(vse,AT_ENTER,o);
		if( !vs_test(t->vs[AT_MEMBER],vse)
		     && !vs_test(t->recursive_vs[AT_MEMBER],vse)
			)
			perm=0;
	}
	if( perm )
	{	*((u_int32_t*)(o->data+t->type->cinfo_offset))=(u_int32_t)(t);
		object_do_sethandler(o);
		*((u_int32_t*)(ret->data))=(u_int32_t)(1);
	}
	else
	{
		*((u_int32_t*)(ret->data))=(u_int32_t)(0);
	}
	return;
}

int cmds_init( void )
{
	lex_addkeyword("constable_pid",Tbuildin,(unsigned long)cmd_constable_pid);
	lex_addkeyword("primaryspace",Tbuildin,(unsigned long)cmd_primaryspace);
	lex_addkeyword("enter",Tbuildin,(unsigned long)cmd_enter);
	lex_addkeyword("hex",Tbuildin,(unsigned long)cmd_hex);
//	lex_addkeyword("log",Tbuildin,(unsigned long)cmd_log);
	return(0);
}


