
#include "execute.h"
#include "language.h"
#include "variables.h"
#include "../constable.h"
#include "../comm/comm.h"

int force_do( u_int32_t fc, struct event_context_s *c, int(*getarg)(struct register_s*) );

static u_int32_t *stack=NULL;
static int stack_size;
static int stack_pos;
static int base_stack_pos=0;

/* stack format:
stack_pos->
		local data
		....
base_stack_pos->local_vars
		base_stack_pos
        	return address
		nr_vars
		arguments
		.....
		.....
		local_vars
		base_stack_pos
		------BEGIN------ 
 */

#define	LTI	0
#define	LTS	1
#define	LTP	2
#define	LTO	3
#define	LTa	4

static void push( u_int32_t data )
{
	if( stack_pos>=stack_size )
	{	stack_size= stack_pos + 256;
		if( (stack=realloc(stack,stack_size*sizeof(u_int32_t)))==NULL )
		{	fatal("Stack overrun");
		}
	}
	stack[stack_pos++]=data;
}

static u_int32_t pop( void )
{
	if( stack_pos <= 0 )
	{	fatal("Stack underrun");
	}
	return( stack[--stack_pos] );
}

struct medusa_attribute_s execute_attr_int={ 0,4,MED_TYPE_SIGNED|MED_TYPE_READ_ONLY,"" };
struct medusa_attribute_s execute_attr_str={ 0,MAX_REG_SIZE,MED_TYPE_STRING|MED_TYPE_READ_ONLY,"" };
struct medusa_attribute_s execute_attr_pointer={ 0,4,MED_TYPE_UNSIGNED|MED_TYPE_READ_ONLY,"" };

void obj_to_reg( struct register_s *r, struct object_s *o, char *attr )
{
	/* init register */
	r->flags=o->flags;
	r->object=NULL;
	r->class=NULL;
	if( attr==NULL )
	{	r->attr=&(o->attr);
		r->object=o;
		r->class=o->class;
		r->data=o->data;
		return;
	}
	if( (r->attr=get_attribute(o->class,attr))==NULL )
	{	runtime("Variable %s has no member %s",o->attr.name,attr);
		r->attr=&execute_attr_int;
		r->data=r->buf;
		*((u_int32_t*)(r->buf))=0;
		return;
	}
	r->data= o->data;
	return;
}

void r_pop( struct register_s *r )
{ u_int32_t x;
  int i,n;
	/* init register */
	r->flags=OBJECT_FLAG_LOCAL;
	r->object=NULL;
	r->class=NULL;
	x=pop();
	switch( x )
	{	case LTI:
			r->attr=&execute_attr_int;
			r->data=r->buf;
			*((u_int32_t*)(r->data))=pop();
			return;
		case LTS:
//			r->attr=&execute_attr_str;
			r->tmp_attr=execute_attr_str;
			r->data=(char *)pop();
			r->tmp_attr.length=strlen( (char *)(r->data) )+1;
			r->attr=&(r->tmp_attr);
			return;
		case LTP:
			r->attr=&execute_attr_pointer;
			r->data=r->buf;
			*((u_int32_t*)(r->data))=pop();
			return;
		case LTO:
			obj_to_reg(r,(struct object_s*)(pop()),NULL);
			return;
		default:
			if( x==LTa )
			{	r->attr=&(r->tmp_attr);
				for(i=0;i<(sizeof(r->tmp_attr)+3)>>2;i++)
					((u_int32_t*)(r->attr))[i]=pop();
			}
			else
				r->attr=(struct medusa_attribute_s *)x;
			if( r->attr->type==MED_TYPE_END )
				r->class=(struct class_s*)(pop());
			x=pop();
			if( x!=0 )
			{	r->data=(char*)x;
				return;
			}
			r->data=r->buf;
			n= (r->attr->length + 3) >> 2;
			for(i=0;i<n;i++)
				((u_int32_t*)(r->data))[i]=pop();
			return;
	}
}

void r_push( struct register_s *r )
{ int n;
	if( r->attr==&execute_attr_int )
	{	push( *((u_int32_t*)(r->data)) );
		push( LTI );
	}
	else if( r->attr==&execute_attr_pointer )
	{	push( *((u_int32_t*)(r->data)) );
		push( LTP );
	}
	else if( r->data!=r->buf && r->object!=NULL )
	{	push( (u_int32_t)(r->object) );
		push( LTO );
	}
	else if( r->attr==&execute_attr_str && r->data!=r->buf )
	{	push( (u_int32_t)(r->data) );
		push( LTS );
	}
	else
	{	if( r->data==r->buf )
		{	n= (r->attr->length + 3) >> 2;
			for(n--;n>=0;n--)
				push( ((u_int32_t*)(r->data))[n] );
			push( 0 );
		}
		else
			push( (u_int32_t)(r->data) );
		if( r->attr->type==MED_TYPE_END )
			push((u_int32_t)(r->class));
		if( r->attr==&(r->tmp_attr) )
		{	for(n=((sizeof(r->tmp_attr)+3)>>2)-1;n>=0;n--)
				push( ((u_int32_t*)(r->attr))[n] );
			push( LTa );
		}
		else
			push( (u_int32_t)(r->attr) );
	}
}

static int fn_getsval_p=0;
static int fn_getsval( struct register_s *r )
{ int z_stack_pos,i,n;
	z_stack_pos=stack_pos;
	n=pop();
	if( r==NULL )
	{	stack_pos=z_stack_pos;
		return(n);
	}
	if( fn_getsval_p>=n )
	{	stack_pos=z_stack_pos;
		return(0);
	}
	for(i=n-fn_getsval_p-1;i>=0;i--)
	{	r_pop(r);
	}
	stack_pos=z_stack_pos;
	fn_getsval_p++;
	if( r->object==NULL )	/* lebo buildin funkcie potrebuju kobjekty */
		r_imm(r);
	return(1);
}

static int fn_getargsval( struct register_s *r, int i )
{ int z_stack_pos,n;
	if( i<=0 )
		return(0);
	z_stack_pos=stack_pos;
	stack_pos=base_stack_pos;
	pop();
	pop();
	n=pop();
	i=n-i;
	if( i<0 )
	{	stack_pos=z_stack_pos;
		return(0);
	}
	for(;i>=0;i--)
	{	r_pop(r);
	}
	stack_pos=z_stack_pos;
	r_imm(r);
	return(1);
}

static struct object_s *data_find_var( struct event_context_s *c, char *name )
{ struct object_s *o;

	if( c->operation.class!=NULL
	    && !strncmp(c->operation.attr.name,name,MEDUSA_ATTRNAME_MAX) )
		return(&(c->operation));
	else if( c->subject.class!=NULL
	    && !strncmp(c->subject.attr.name,name,MEDUSA_ATTRNAME_MAX) )
		return(&(c->subject));
	else if( c->object.class!=NULL
	    && !strncmp(c->object.attr.name,name,MEDUSA_ATTRNAME_MAX) )
		return(&(c->object));
	else
	{ u_int32_t p;
#ifdef LOCAL_VARS_RECURSIVE_SEARCH
		for(p=base_stack_pos;p!=0;p=stack[p-1])
#else
		p=base_stack_pos;
#endif
		   if( (o=get_var((struct object_s*)(stack[p]),name))!=NULL )
			return(o);
	}
	if( (o=get_var(c->local_vars,name))!=NULL )
		return(o);
	runtime("Undefined variable %s",name);
	return(NULL);
}

void reg_load_var( struct register_s *r, struct event_context_s *c, char *name, char *attr )
{ struct object_s *o=NULL;

	/* init register */
	r->flags=0;
	r->object=NULL;
	r->class=NULL;
	if( attr==NULL )
	{	if( c->operation.class!=NULL
		    && (r->attr=get_attribute(c->operation.class,name))!=NULL )
		{	r->data=c->operation.data;
			r->flags=c->operation.flags;
			return;
		}
		if( c->subject.class!=NULL
		    && (r->attr=get_attribute(c->subject.class,name))!=NULL )
		{	r->data=c->subject.data;
			r->flags=c->subject.flags;
			return;
		}
		if( c->object.class!=NULL
		    && (r->attr=get_attribute(c->object.class,name))!=NULL )
		{	r->data=c->object.data;
			r->flags=c->object.flags;
			return;
		}
	}
	if( (o=data_find_var(c,name))==NULL )
	{	r->attr=&execute_attr_int;
		r->data=r->buf;
		*((u_int32_t*)(r->buf))=0;
		return;
	}
	obj_to_reg(r,o,attr);
}

static struct register_s r0,r1;

char *execute_get_last_data( void )
{
	return(r0.data);
}

struct medusa_attribute_s *execute_get_last_attr( void )
{
	return(r0.attr);
}

int execute_handler( struct event_handler_s *h, struct event_context_s *c )
{ u_int32_t *p,cmd=oNOP,x;
  struct object_s *v;
  u_int32_t stack_start;

	c->result=RESULT_OK;

	p=(u_int32_t *)(h->data);

	stack_start=stack_pos;
	push(base_stack_pos);
	base_stack_pos=stack_pos;
	push(0);	/* local_vars */
for(;;)
{	cmd=*p++;
	switch( cmd )
	{
	case oNOP:	break;
	case oLTI:	push(LTI);
			break;
	case oLTS:	push(LTS);
			break;
	case oLTP:	push(LTP);
			break;
	case oLDI:	push(*p++);
			break;
	case oLD0:	push(0);
			push(LTI);
			break;
	case oLD1:	push(1);
			push(LTI);
			break;
	case oDUP:	r_pop(&r0);
			r_push(&r0);
			r_push(&r0);
			break;
	case oSWP:	r_pop(&r0);
			r_pop(&r1);
			r_push(&r0);
			r_push(&r1);
			break;
	case oDEL:	r_pop(&r0);
			break;
	case oDEn:	r_pop(&r1);
			x=pop();
			while( x )
			{	r_pop(&r0);
				x--;
			}
			r_push(&r1);
			break;
	case oIMM:	r_pop(&r0);
			r_imm(&r0);
			r_push(&r0);
			break;
	case oSTO:	r_pop(&r0);
			r_pop(&r1);
			r_sto(&r1,&r0);
			r_push(&r1);
			break;
	case oFET:	if( (v=data_find_var(c,(char *)(*p++)))!=NULL )
			{	if( comm_fetch_object(v)<0 )
				{	runtime("Can't fetch object");
					push(0);
				}
				else	push(1);
				push(LTI);
			}
			else
			{	push(0);
				push(LTI);
			}
			break;
	case oUPD:	if( (v=data_find_var(c,(char *)(*p++)))!=NULL )
			{	if( comm_update_object(v)<0 )
				{	runtime("Can't update object");
					push(0);
				}
				else	push(1);
				push(LTI);
			}
			else
			{	push(0);
				push(LTI);
			}
			break;
	case oLDC:	x=*p++;
			load_constant(&r0,x,(char*)(*p++));
			r_push(&r0);
			break;
	case oLDV:	reg_load_var(&r0,c,(char*)(*p++),NULL);
			r_push(&r0);
			break;
	case oLDS:	x=*p++;
			reg_load_var(&r0,c,(char*)x,(char*)(*p++));
			r_push(&r0);
			break;
	case oNEW:	x=*p++;
			if( (v=alloc_var((char*)(*p++),NULL,(struct class_s*)x))!=NULL )
				var_link((struct object_s**)(&(stack[base_stack_pos])),v);
			else runtime("Can't allocate varible",((struct class_s*)x)->m.name);
			obj_to_reg(&r0,v,NULL);
			r_push(&r0);
			break;
	case oTNW:	x=*p++;
			if( (v=alloc_var((char*)(*p++),NULL,(struct class_s*)x))!=NULL )
				var_link(&(c->local_vars),v);
			else runtime("Can't allocate varible",((struct class_s*)x)->m.name);
			obj_to_reg(&r0,v,NULL);
			r_push(&r0);
			break;
	case oADD:
	case oSUB:
	case oMUL:
	case oDIV:
	case oMOD:
	case oSHL:
	case oSHR:
	case oLT:
	case oGT:
	case oLE:
	case oGE:
	case oEQ:
	case oNE:
	case oOR:
	case oAND:
	case oXOR:
			r_pop(&r0);
			r_pop(&r1);
			do_bin_op(cmd,&r1,&r0);
			r_push(&r1);
			break;
	case oNOT:	r_pop(&r0);
			r_not(&r0);
			r_push(&r0);
			break;
	case oNEG:	r_pop(&r0);
			r_imm(&r0);
			r_neg(&r0);
			r_push(&r0);
			break;

	case oJR:	p+= *p+1;	/* p+= *p++; */
			break;
	case oJIZ:	r_pop(&r0);
			if( ! r_nz(&r0) )
				p+= *p+1; /* p+= *p++; */
			else	p++;
			break;
	case oJNZ:	r_pop(&r0);
			if( r_nz(&r0) )
				p+= *p+1; /* p+= *p++; */
			else	p++;
			break;
	case oJSR:	x=*p++;
//printf("%p JSR %x\n",p-2,x);
			if( x==0 )
			{	runtime("Call undefined function");
				push(0);
				push(LTI);
				break;
			}
			push((u_int32_t)p);
			push(base_stack_pos);
			base_stack_pos=stack_pos;
			push(0);	/* local_vars */
			p=(u_int32_t*)x;
			break;
	case oRET:	
//printf("%p RET\n",p-1);
			r_pop(&r0);
			r_imm(&r0);	/* POZOR! mohla by byt z local_vars */
			free_vars((struct object_s**)(&(stack[base_stack_pos])));
			pop(); /* local_vars */
			base_stack_pos=pop();
			if( stack_pos <= stack_start )
			{	
				c->result=r_int(&r0);
				stack_pos=stack_start;
				free_vars(&(c->local_vars));
				return( c->result );
			}
			
			p= (u_int32_t*) pop();
			r_push(&r0);
			break;
	case oARG:	x=*p++;
			if( !fn_getargsval(&r0,x) )
				runtime("Invalid argument $%d",x);
			r_push(&r0);
			break;
	case oBIN:	x=*p++;
			fn_getsval_p=0;
			/* init register */
			r0.flags=0;
			r0.object=NULL;
			r0.class=NULL;
			r0.attr=&execute_attr_int; r0.data=r0.buf;
			*((u_int32_t*)(r0.data))= -1;
			((void(*)(struct event_context_s*,struct register_s*,int(*)(struct register_s*)))(x))(c,&r0,fn_getsval);
			r_push(&r0);
			break;
	case oFRC:	x=*p++;
			fn_getsval_p=0;
			push( /* force_do(x,c,fn_getsval)*/ 0 );
			push(LTI);
			break;
#ifndef DEBUG_TRACE
	default:	runtime("Illegal instruction 0x%08x 0x%08x 0x%08x",cmd,p[0],p[1]);
#else
	default:	runtime("Illegal instruction [%s] 0x%08x 0x%08x 0x%08x",h->op_name+DT_POS_MAX,cmd,p[0],p[1]);
#endif
	}
}
}

int execute_init( void )
{
	stack_size=256;
	stack_pos=0;
	if( (stack=malloc(stack_size*sizeof(u_int32_t)))==NULL )
		return(init_error(Out_of_memory));
	return(0);
}

