
#include "rbac.h"
#include "../space.h"
#include "../language/error.h"
#include "../language/language.h"
#include <stdlib.h>

/*
	- vytvorit rolu			create <role>
	- hirerchicky usporiadat role	hierarchy <suprole> <subrole>
	- priradit uzivatela		user <role> <user>
	- pridat pravo			permission <role> <access> <space>
 */

char *Permission_denied="Permission denied";

int load_users( char *file );

int rbac_adm_create_role( struct object_s *subject, char *name )
{ vs_t vsp[MAX_VS_BITS/32];
  struct role_s *r;
	subject_get_vs(vsp,AT_CREATE,subject);
	if( !vs_test(r_role->vs,vsp) && !vs_test(r_role->recursive_vs,vsp) &&
	    !vs_test(r_ROLE->vs,vsp) && !vs_test(r_ROLE->recursive_vs,vsp) )
	{	errstr=Permission_denied;
		return(-1);
	}
	if( (r=rbac_role_add(name))==NULL )
		return(-1);
	return(0);
}

int rbac_adm_delete_role( struct object_s *subject, char *name )
{ vs_t vsp[MAX_VS_BITS/32];
  vs_t vsr[MAX_VS_BITS/32];
  struct role_s *r;
	if( (r=rbac_role_find(name))==NULL )
	{	errstr="Undefined role";
		return(-1);
	}
	subject_get_vs(vsp,AT_ERASE,subject);
	if( !vs_test(r_role->vs,vsp) && !vs_test(r_role->recursive_vs,vsp) &&
	    !vs_test(r_ROLE->vs,vsp) && !vs_test(r_ROLE->recursive_vs,vsp) )
	{	errstr=Permission_denied;
		return(-1);
	}
	rbac_get_role_vs(vsr,r);
	if( !vs_test(vsr,vsp) )
	{	errstr=Permission_denied;
		return(-1);
	}
	return( rbac_role_del(r) );
}

int rbac_adm_hierarchy( int add, struct object_s *subject, char *sup_name, char *sub_name )
{ vs_t vsp[MAX_VS_BITS/32];
  vs_t vsr[MAX_VS_BITS/32];
  struct role_s *rp,*rb;
  	if( (rp=rbac_role_find(sup_name))==NULL
	    || (rb=rbac_role_find(sub_name))==NULL )
	{	errstr="Undefined role";
		return(-1);
	}
	if( add )
		subject_get_vs(vsp,AT_CREATE,subject);
	else
		subject_get_vs(vsp,AT_ERASE,subject);

	if( !vs_test(r_role->recursive_vs,vsp) && !vs_test(r_ROLE->recursive_vs,vsp) )
	{	
		rbac_get_role_vs(vsr,rb);
		if( !vs_test(vsr,vsp) )
		{	errstr=Permission_denied;
			return(-1);
		}
		rbac_get_role_vs(vsr,rp);
		if( !vs_test(vsr,vsp) )
		{	errstr=Permission_denied;
			return(-1);
		}
	}
	if( add )
	{	if( rbac_set_hierarchy(rp,rb)<0 )
			return(-1);
	}
	else
	{	if( rbac_del_hierarchy(rp,rb)<0 )
			return(-1);
	}
	return(0);
}

int rbac_adm_user( int add, struct object_s *subject, char *role, char *user )
{ vs_t vsp[MAX_VS_BITS/32];
  vs_t vsr[MAX_VS_BITS/32];
  struct role_s *r;
  struct user_s *u;
  	if( (r=rbac_role_find(role))==NULL )
	{	errstr="Undefined role";
		return(-1);
	}
	if( (u=rbac_user_find(user))==NULL )
	{	errstr="Undefined user";
		return(-1);
	}
	subject_get_vs(vsp,AT_CONTROL,subject);

	if( !vs_test(r_role->recursive_vs,vsp) && !vs_test(r_ROLE->recursive_vs,vsp) )
	{	
		rbac_get_role_vs(vsr,r);
		if( !vs_test(vsr,vsp) )
		{	errstr=Permission_denied;
			return(-1);
		}
	}
	if( !vs_test(r_user->recursive_vs,vsp) )
	{	rbac_get_user_vs(vsr,u);
		if( !vs_test(vsr,vsp) )
		{	errstr=Permission_denied;
			return(-1);
		}
	}
	if( add )
	{	if( rbac_add_ua(u,r)<0 )
			return(-1);
	}
	else
	{	if( rbac_del_ua(u,r)<0 )
			return(-1);
	}
	return(0);
}

static int str2at( char *str )
{ lextab_t *l=keywords;
	while(l->keyword!=NULL)
	{	if( l->sym==Taccess && !strcmp(str,l->keyword) )
		return(l->data);
		l++;
	}
	return(-1);
}

int rbac_adm_perm( int add, struct object_s *subject, char *role, char *access, char *space )
{ vs_t vsp[MAX_VS_BITS/32];
  vs_t vsr[MAX_VS_BITS/32];
  struct role_s *r;
  struct space_s *t;
  struct rbac_perm_s *p;
  int which;
  	which=str2at(access);
  	if( which<=0 || which>=NR_ACCESS_TYPES )
	{	errstr="Invalid access type";
		return(-1);
	}
  	if( (r=rbac_role_find(role))==NULL )
	{	errstr="Undefined role";
		return(-1);
	}
	if( (t=space_find(space))==NULL )
	{	errstr="Undefined space";
		return(-1);
	}
	subject_get_vs(vsp,AT_CONTROL,subject);

	if( !vs_test(r_role->recursive_vs,vsp) && !vs_test(r_ROLE->recursive_vs,vsp) )
	{	rbac_get_role_vs(vsr,r);
		if( !vs_test(vsr,vsp) )
		{	errstr=Permission_denied;
			return(-1);
		}
	}
	if( (p=(struct rbac_perm_s *)find_one((struct tree_s *)r_perm,t->name))
		==NULL )
	{	errstr=Permission_denied;
		return(-1);
	}
	if( !vs_test(p->vs,vsp) )
	{	if( (p=(struct rbac_perm_s *)find_one((struct tree_s *)p,access))==NULL || !vs_test(p->vs,vsp) )
		{	errstr=Permission_denied;
			return(-1);
		}
	}
	
	if( add )
	{	if( rbac_role_add_perm(r,which,t)<0 )
			return(-1);
	}
	else
	{	if( rbac_role_del_perm(r,which,t)<0 )
			return(-1);
	}
	return(0);
}

int rbac_adm( struct object_s *subject, char *op, char *s1, char *s2, char *s3 )
{ int x;
	if( !strcmp(op,"lu") )
		return( load_users("/etc/passwd") );
	if( !strcmp(op,"cr") )
		return( rbac_adm_create_role(subject,s1) );
	if( !strcmp(op,"dr") )
		return( rbac_adm_delete_role(subject,s1) );
	if( !strcmp(op,"ra") )
		return( rbac_adm_hierarchy(1,subject,s1,s2) );
	if( !strcmp(op,"rd") )
		return( rbac_adm_hierarchy(0,subject,s1,s2) );
	if( !strcmp(op,"ua") )
		return( rbac_adm_user(1,subject,s1,s2) );
	if( !strcmp(op,"ud") )
		return( rbac_adm_user(0,subject,s1,s2) );
	if( !strcmp(op,"pa") )
		return( rbac_adm_perm(1,subject,s1,s2,s3) );
	if( !strcmp(op,"pd") )
		return( rbac_adm_perm(0,subject,s1,s2,s3) );
	if( !strcmp(op,"save") )
	{	/*if( (x=atoi(s2))==0 )*/	x=16;
		if( rbac_save(rbac_conffile /*s1*/,x)<0 )
		{	errstr="Can't write RBAC configuration file";
			return(-1);
		}
		return(0);
	}
	errstr="Unknown command";
	return(-1);
}
