
#include "rbac.h"
#include "../medusa_object.h"

static struct medusa_comm_class_s rbac_user_mclass={
	0,sizeof(struct user_s),"rbac_user"
};

static struct medusa_comm_attribute_s rbac_user_attr[]={
	MED_ATTR_x(struct user_s,uid,"uid",MED_TYPE_UNSIGNED|MED_TYPE_PRIMARY_KEY|MED_TYPE_READ_ONLY),
	MED_ATTR_BITMAP(struct user_s,vs,"vs"),
	MED_ATTR_BITMAP(struct user_s,cinfo,"o_cinfo"),
	MED_ATTR_STRING(struct user_s,name,"name"),
	MED_ATTR_END
};

static struct medusa_comm_class_s rbac_role_mclass={
	0,sizeof(struct role_s),"rbac_role"
};

static struct medusa_comm_attribute_s rbac_role_attr[]={
	MED_ATTR_BITMAP(struct role_s,vs_sub,"vs"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_RECEIVE],"vsr"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_SEND],"vsw"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_SEE],"vss"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_CREATE],"vsc"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_ERASE],"vsd"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_ENTER],"vse"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_CONTROL],"vsx"),
	MED_ATTR_BITMAP(struct role_s,cinfo_sub,"o_cinfo"),
	MED_ATTR_x(struct role_s,name,"name",MED_TYPE_STRING|MED_TYPE_PRIMARY_KEY|MED_TYPE_READ_ONLY),
	MED_ATTR_END
};

static struct medusa_comm_class_s rbac_ROLE_mclass={
	0,sizeof(struct role_s),"rbac_ROLE"
};

static struct medusa_comm_attribute_s rbac_ROLE_attr[]={
	MED_ATTR_BITMAP(struct role_s,vs[AT_MEMBER],"vs"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_RECEIVE],"vsr"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_SEND],"vsw"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_SEE],"vss"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_CREATE],"vsc"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_ERASE],"vsd"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_ENTER],"vse"),
	MED_ATTR_BITMAP(struct role_s,vs[AT_CONTROL],"vsx"),
	MED_ATTR_BITMAP(struct role_s,cinfo_sup,"o_cinfo"),
	MED_ATTR_x(struct role_s,name,"name",MED_TYPE_STRING|MED_TYPE_PRIMARY_KEY|MED_TYPE_READ_ONLY),
	MED_ATTR_END
};

int rbac_set_object( struct object_s *o, u_int32_t *cinfo, void *user )
{
  struct tree_s *t;
  int a;

	if( (t=(struct tree_s *)(*cinfo))==NULL )
	{
		t=(struct tree_s*)user;
		t=find_one(t,o->data + o->class->name_offset);
		*cinfo=(u_int32_t)t;
	}
#ifdef USE_ALT
	for(;t!=NULL;t=t->alt)
	{
#endif
		for(a=0;a<NR_ACCESS_TYPES;a++)
		{	object_add_vs(o,a,t->vs[a]);
			object_add_vs(o,a,t->recursive_vs[a]);
		}
		object_add_event(o,&(t->event[0]));
		object_add_event(o,&(t->recursive_event[0]));
#ifdef USE_ALT
	}
#endif
	return(0);
}

static int rbac_set_roles( struct object_s *o, u_int32_t *cinfo, void *user )
{
	VALIDATE_ROLES();
	return(0);
}

int rbac_object_init( void )
{
  static struct tree_type_s t_user,t_role,t_ROLE;
	if( (rbac_user_class=add_class(&rbac_user_mclass,&(rbac_user_attr[0])))==NULL )
		return(init_error("rbac: Can't register user class"));
	if( (rbac_role_class=add_class(&rbac_role_mclass,&(rbac_role_attr[0])))==NULL )
		return(init_error("rbac: Can't register role class"));
	if( (rbac_ROLE_class=add_class(&rbac_ROLE_mclass,&(rbac_ROLE_attr[0])))==NULL )
		return(init_error("rbac: Can't register ROLE class"));
	

	if( (rbac_user_cinfo=class_alloc_object_cinfo(rbac_user_class))<0 )
		return(init_error("rbac: Can't allocate user cinfo"));
	if( (rbac_role_cinfo=class_alloc_object_cinfo(rbac_role_class))<0 )
		return(init_error("rbac: Can't allocate role cinfo"));
	if( (rbac_ROLE_cinfo=class_alloc_object_cinfo(rbac_ROLE_class))<0 )
		return(init_error("rbac: Can't allocate ROLE cinfo"));

	t_user.name="user";
	t_user.size=sizeof(struct tree_s);
	t_user.obj_class=rbac_user_class;
	t_user.cinfo_offset=rbac_user_cinfo;
	t_user.user=NULL;
	t_user.init=NULL;
	t_user.child_type=&t_user;
	if( (rbac_user_root=register_tree_type(&t_user))==NULL )
		return(init_error("rbac: Can't register user tree type"));
	
	t_role.name="role";
	t_role.size=sizeof(struct tree_s);
	t_role.obj_class=rbac_role_class;
	t_role.cinfo_offset=rbac_role_cinfo;
	t_role.user=NULL;
	t_role.init=NULL;
	t_role.child_type=&t_role;
	if( (rbac_role_root=register_tree_type(&t_role))==NULL )
		return(init_error("rbac: Can't register role tree type"));
	
	t_ROLE.name="ROLE";
	t_ROLE.size=sizeof(struct tree_s);
	t_ROLE.obj_class=rbac_ROLE_class;
	t_ROLE.cinfo_offset=rbac_ROLE_cinfo;
	t_ROLE.user=NULL;
	t_ROLE.init=NULL;
	t_ROLE.child_type=&t_ROLE;
	if( (rbac_ROLE_root=register_tree_type(&t_ROLE))==NULL )
		return(init_error("rbac: Can't register ROLE tree type"));
	
	if( class_add_sethandler(rbac_user_class,rbac_set_object,rbac_user_cinfo,rbac_user_root)<0 )
		return(init_error("rbac: Can't register user rethandler"));
	if( class_add_sethandler(rbac_role_class,rbac_set_roles,rbac_role_cinfo,rbac_role_root)<0 )
		return(init_error("rbac: Can't register role rethandler"));
	if( class_add_sethandler(rbac_ROLE_class,rbac_set_roles,rbac_ROLE_cinfo,rbac_ROLE_root)<0 )
		return(init_error("rbac: Can't register ROLE rethandler"));

	return(0);
}

