

static struct medusa_attribute_s *attr_uid;
static struct event_type_s *event_setuid;
static int session_cinfo_offset;

static int rbac_set_proc( struct object_s *o, u_int32_t *cinfo, void *user )
{ struct user_s *u;
  struct role_s *r;
  uid_t uid;
  u_int32_t session;
  int i,a;

	VALIDATE_ROLES();
	object_get_val(o,attr_uid,&uid,sizeof(uid));
	if( (u=rbac_user_find_by_uid(uid))==NULL )
	{	runtime("RBAC: Unknown uid %d",uid);
		return(-1);
	}
	session=*cinfo;
	i=0;
	while( session && i<USER_MAX_ROLES )
	{	if( (session & 0x01) && (r=u->roles[i]) )
		{	for(a=0;a<NR_ACCESS_TYPES;a++)
				object_add_vs(o,a,r->vs[a]);
			object_add_vs(o,AT_MEMBER,r->vs_sub);
		}
		session>>=1;
		i++;
	}
	object_add_event(o,event_setuid->mask);
	return(0);
}

static int rbac_proc_getvs( vs_t *vs, int n, struct object_s *o, u_int32_t *cinfo )
{ struct user_s *u;
  struct role_s *r;
  uid_t uid;
  u_int32_t session;
  int i;

	VALIDATE_ROLES();
	object_get_val(o,attr_uid,&uid,sizeof(uid));
	if( (u=rbac_user_find_by_uid(uid))==NULL )
	{	runtime("RBAC: Unknown uid %d",uid);
		return(-1);
	}
	session=*cinfo;
	i=0;
	while( session && i<USER_MAX_ROLES )
	{	if( (session & 0x01) && (r=u->roles[i]) )
		{	vs_add(r->vs[n],vs);
//			if( n==AT_MEMBER )
//				vs_add(r->vs_sub,vs);
		}
		session>>=1;
		i++;
	}
	return(0);
}

static int rbac_setuid_handler( struct event_handler_s *h, struct event_context_s *c )
{
	object_do_sethandler(&(c->subject));
	c->result=RESULT_OK;
	return(0);
}

int rbac_init()
{ struct class_s *c;
  struct event_handler_s *eh;
	if( rbac_object_init()<0 )
		return(-1);
	if( (event_setuid=event_type_find_name("setuid"))==NULL )
		return(init_error("rbac: Operation setuid does not exist!"));
	if( (c=e->op[0])==NULL || e->object!=e->op[0] )
		return(init_error("rbac: Invalid operation setuid"));
	if( (attr_uid=get_attribute(c,"uid"))==NULL )
		return(init_error("rbac: Can't find attribute uid"));
	if( (session_cinfo_offset=class_alloc_subject_cinfo(c))<0 )
		if( (session_cinfo_offset=class_alloc_object_cinfo(c))<0 )
			return(init_error("rbac: Can't allocate cinfo for session"));
	if( class_add_sethandler(c,rbac_set_proc,session_cinfo_offset,NULL)<0 )	
		return(init_error("rbac: Can't add sethandler to process"));
	if( class_add_getvs(c,rbac_proc_getvs,session_cinfo_offset)<0 )
		return(init_error("rbac: Can't add getvs handler to process"));
	if( (eh=malloc(sizeof(struct event_handler_s)))==NULL )
		return(init_error(Out_of_memory));
	strncpy(eh->op_name,"rbac:setuid",MEDUSA_OPNAME_MAX-1);
	eh->handler=rbac_setuid_handler;
	eh->local_vars=NULL;
	if( register_event_handler(h,event_setuid,ALL_OBJ,ALL_OBJ)<0 )
		return(init_error("rbac: Can't register setuid handler"));
	return(0);
}

