
#include "constable.h"
#include "space.h"

struct space_s *global_spaces=NULL;

static ltree_t *new_path( ltree_t *prev, struct tree_s *path )
{ ltree_t *l;
	if( (l=malloc(sizeof(ltree_t)))==NULL )
	{	errstr=Out_of_memory;
		return(NULL);
	}
	l->prev=prev;
	l->tree=path;
	return(l);
}

static levent_t *new_levent( levent_t **prev, struct event_handler_s *handler, struct space_s *subject, struct space_s *object )
{ levent_t *l;
	if( (l=malloc(sizeof(levent_t)))==NULL )
	{	errstr=Out_of_memory;
		return(NULL);
	}
	l->handler=handler;
	l->subject=subject;
	l->object=object;
	l->prev=*prev;
	*prev=l;
	return(l);
}

struct space_s *space_find( char *name )
{ struct space_s *t;
	for(t=global_spaces;t!=NULL;t=t->next)
		if( !strcmp(t->name,name) )
			return(t);
	return(NULL);
}

struct space_s *space_create( char *name, int primary )
{ struct space_s *t;
  int a;
	if( name!=NULL )
	{	if( space_find(name)!=NULL )
		{	errstr="Redefinition of space";
			return(NULL);
		}
	}
	else	name=" _";
	if( (t=malloc(sizeof(struct space_s)+strlen(name)+1))==NULL )
	{	errstr=Out_of_memory;
		return(NULL);
	}
	strcpy(t->name,name);
	t->my_vs=NULL;
	for(a=0;a<NR_ACCESS_TYPES;a++)
		vs_clear(t->vs[a]);
	t->levent=NULL;
	t->ltree=NULL;
	t->ltree_exclude=NULL;
	t->primary=primary;
	t->next=global_spaces;
	global_spaces=t;
	return(t);
}

static int is_excluded( struct tree_s *t, struct space_s *space )
{ ltree_t *e;
	for(e=space->ltree_exclude;e!=NULL;e=e->prev)
	{	if( t == e->tree )
			return(1);
	}
	return(0);
}
static int is_excluded_recursive( struct tree_s *t, struct space_s *space )
{ ltree_t *e;
	for(e=space->ltree_exclude;e!=NULL;e=e->prev)
	{	if( t == e->tree && e->recursive )
			return(1);
	}
	return(0);
}

static void recursive_set_primary_space( struct tree_s *t, struct space_s *space )
{ struct tree_s *p;
	if( is_excluded_recursive(t,space) )
		return;
	if( t->recursive_primary_space!=NULL )
		warning("Redefinition of recursive primary space %s to %s",
			t->recursive_primary_space->name,
			space->name);
	if( !is_excluded(t,space) )
		t->recursive_primary_space=space;
	for(p=t->child;p!=NULL;p=p->next)
		recursive_set_primary_space(p,space);
	for(p=t->reg;p!=NULL;p=p->next)
		recursive_set_primary_space(p,space);
}

static void set_primary_space( struct tree_s *t, int recursive, struct space_s *space )
{
	if( recursive )
		recursive_set_primary_space(t,space);
	else
	{	if( t->primary_space!=NULL )
			warning("Redefinition of primary space %s to %s",
				t->primary_space->name,
				space->name);
		t->primary_space=space;
	}
}

static void recursive_unset_primary_space( struct tree_s *t, struct space_s *space )
{ struct tree_s *p;
	if( t->recursive_primary_space==space )
		t->recursive_primary_space=NULL;
	for(p=t->child;p!=NULL;p=p->next)
		recursive_set_primary_space(p,space);
	for(p=t->reg;p!=NULL;p=p->next)
		recursive_set_primary_space(p,space);
}

static void unset_primary_space( struct tree_s *t, int recursive, struct space_s *space )
{
	if( recursive )
		recursive_unset_primary_space(t,space);
	else
	{	if( t->primary_space==space )
			t->primary_space=NULL;
	}
}

/* ----------------------------------- */

static void recursive_add_event( struct tree_s *f, struct event_type_s *type, struct space_s *space )
{ struct tree_s *t;
	if( is_excluded_recursive(f,space) )
		return;
	if( !is_excluded(f,space) &&
	    type->object==f->type->obj_class )
		event_mask_or2(f->recursive_event,type->mask);
	for(t=f->child;t!=NULL;t=t->next)
		recursive_add_event(t,type,space);
	for(t=f->reg;t!=NULL;t=t->next)
		recursive_add_event(t,type,space);
}

static int tree_add_event( struct tree_s *p, int recursive, struct event_type_s *type, struct space_s *space )
{
	if( recursive )
		recursive_add_event(p,type,space);
	else if( type->object==p->type->obj_class )
		event_mask_or2(p->event,type->mask);
	return(0);
}

/* ----------------------------------- */

static void recursive_add_vs( struct tree_s *f, int which, const vs_t *vs, struct space_s *space )
{ struct tree_s *t;
	if( is_excluded_recursive(f,space) )
		return;
	if( !is_excluded(f,space) )
		vs_add(vs,f->recursive_vs[which]);
	for(t=f->child;t!=NULL;t=t->next)
		recursive_add_vs(t,which,vs,space);
	for(t=f->reg;t!=NULL;t=t->next)
		recursive_add_vs(t,which,vs,space);
}

static int tree_add_vs( struct tree_s *p, int recursive, int which, const vs_t *vs, struct space_s *space )
{
	if( recursive )
		recursive_add_vs(p,which,vs,space);
	else
		vs_add(vs,p->vs[which]);
	return(0);
}

/* ----------------------------------- */



int space_add_path( struct space_s *space, int recursive, struct tree_s *path )
{ // levent_t *l;
//  int a;
	if( path==NULL )	return(-1);
	if( (space->ltree=new_path(space->ltree,path))==NULL )
		return(-1);
	space->ltree->recursive=recursive;
	if( space->primary )
		set_primary_space(path,recursive,space);

/* toto je uz mozno zbytocne !!! a pozor na registraciu eventov
	for(a=0;a<NR_ACCESS_TYPES;a++)
		tree_add_vs(path,recursive,a,space->vs[a],space);
	for(l=space->levent;l!=NULL;l=l->prev)
	{	tree_add_event(path,recursive,l->handler,space);
	}
*/
	return(0);
}

int space_add_exclude( struct space_s *space, int recursive, struct tree_s *path )
{
//  levent_t *l;
//  int a;
	if( path==NULL )	return(-1);
	if( (space->ltree_exclude=new_path(space->ltree_exclude,path))==NULL )
		return(-1);
	space->ltree_exclude->recursive=recursive;
	if( space->primary )
		unset_primary_space(path,recursive,space);
	
	/* FIXME: cele tieto dynamicke veci opravit */
//	for(a=0;a<NR_ACCESS_TYPES;a++)	/* ZLE!!! */
//		path->type->sub_vs(path,recursive,a,space->vs[a]);
//	for(l=space->levent;l!=NULL;l=l->prev)
//	{	tree_sub_event(path,recursive,l->handler); /* to nie je take iste!!! */
//	}
	return(0);
}

int space_add_space( struct space_s *space, struct space_s *f )
{
/* ......!!!!!!!!!!!!!!!!................. */
return(-1);
}

int space_add_event( struct event_handler_s *handler, struct space_s *subject, struct space_s *object )
{ ltree_t *o,*s;
  struct event_type_s *type;
  int registred=0;
	if( (type=event_type_find_name(handler->op_name))==NULL )
	{	error("Unknown event %s\n",handler->op_name);
		return(-1);
	}
	if( ((type->op[0]==NULL) != (subject==NULL)) ||
	    ((type->op[1]==NULL) != (object==NULL)) )
	{	error("Invalid use of event %s\n",handler->op_name);
		return(-1);
	}
	if( subject!=NULL && subject!=ALL_OBJ && new_levent(&(subject->levent),handler,subject,object)==NULL )
		return(-1);
	if( object!=NULL && object!=ALL_OBJ && new_levent(&(object->levent),handler,subject,object)==NULL )
		return(-1);

	if( subject!=NULL && subject!=ALL_OBJ && type->object==type->op[0] )
	{	for(s=subject->ltree;s!=NULL;s=s->prev)
		{	tree_add_event(s->tree,s->recursive,type,subject);
			registred=1;
		}
	}
	if( object!=NULL && object!=ALL_OBJ && type->object==type->op[1] )
	{	for(o=object->ltree;o!=NULL;o=o->prev)
		{	tree_add_event(o->tree,o->recursive,type,object);
			registred=1;
		}
	}
	register_event_handler(handler,type,space_get_vs(subject),space_get_vs(object));
	if( subject==ALL_OBJ && (object==ALL_OBJ || object==NULL) )
		registred=1;
	if( !registred )
		return(-1);
	return(0);
}

int space_add_vs( struct space_s *space, int which, vs_t *vs )
{ ltree_t *l;
	if( which<0 || which>=NR_ACCESS_TYPES )
		return(-1);
	vs_add(vs,space->vs[which]);
	for(l=space->ltree;l!=NULL;l=l->prev)
		tree_add_vs(l->tree,l->recursive,which,vs,space);
	return(0);
}

vs_t *space_get_vs( struct space_s *space )
{ ltree_t *l;
	if( space==NULL || space==ALL_OBJ)
		return((vs_t*)space);
	if( space->my_vs!=NULL )
		return(space->my_vs->vs);
	if( (space->my_vs=vs_alloc(space->name))==NULL )
		return(NULL);
//printf("ZZZ:space_get_vs %s 0x%08x\n",space->name,*(space->my_vs->vs));
	vs_add(space->my_vs->vs,space->vs[AT_MEMBER]);
	for(l=space->ltree;l!=NULL;l=l->prev)
		tree_add_vs(l->tree,l->recursive,AT_MEMBER,space->vs[AT_MEMBER],space);
	return(space->my_vs->vs);
}

