
#include "constable.h"
#include "vs.h"

static int number_of_vs=MAX_VS_BITS/32;
static struct vs_s vs_tab;

int vs_init( int max_vs )
{ int i;
	if( (max_vs % 32) !=0 )
	{	errstr="Invalid number of vs";
		return(-1);
	}
	vs_tab.next=NULL;
	for(i=0;i<MAX_VS_BITS/32;i++)
		vs_tab.vs[i]=0;
	if( max_vs > MAX_VS_BITS )
	{	number_of_vs=MAX_VS_BITS/32;
		errstr="Too many vs";
		return(1);
	}
	number_of_vs= max_vs/32;
	return(0);
}

/* ---------- vs alloc/free ---------- */
struct vs_s *vs_alloc( char *name )
{ struct vs_s *p;
  int i;
  vs_t x;
	for(p=&vs_tab;p->next!=NULL;p=p->next)
	{	if( name[0]!=' ' && !strcmp(name,p->next->name) )
		{	errstr="Redefinition of vs";
			return(NULL);
		}
	}
	if( (p->next=malloc(sizeof(struct vs_s)+strlen(name)+1))==NULL )
	{		errstr=Out_of_memory;
			return(NULL);
	}
	p->next->next=NULL;
	strcpy(p->next->name,name);
	x=0;
	if( p == &vs_tab )	x=1;
	for(i=0;i<MAX_VS_BITS/32;i++)
	{	p->next->vs[i]= (p->vs[i]<<1) | x ;
		x=(p->vs[i])>>31;
	}
	if( x )
	{	free(p->next);
		p->next=NULL;
		errstr="Out of vs";
		return(NULL);
	}
	return(p->next);
}

struct vs_s *vs_find( char *name )
{ struct vs_s *p;
	for(p=vs_tab.next;p!=NULL;p=p->next)
	{	if( !strcmp(name,p->name) )
			return(p);
	}
	errstr="Undefined vs";
	return(NULL);
}

void vs_clear( vs_t *to )
{ int i;
	for(i=0;i<number_of_vs;i++)
		to[i]= 0;
}

void vs_fill( vs_t *to )
{ int i;
	for(i=0;i<number_of_vs;i++)
		to[i]= ~((vs_t)0);
}

void vs_invert( vs_t *to )
{ int i;
	for(i=0;i<number_of_vs;i++)
		to[i]= ~(to[i]);
}

void vs_set( const vs_t *from, vs_t *to )
{ int i;
	for(i=0;i<number_of_vs;i++)
		to[i]= from[i];
}

void vs_add( const vs_t *from, vs_t *to )
{ int i;
	for(i=0;i<number_of_vs;i++)
		to[i]|= from[i];
}

void vs_sub( const vs_t *from, vs_t *to )
{ int i;
	for(i=0;i<number_of_vs;i++)
		to[i]&= ~(from[i]);
}

void vs_mask( const vs_t *from, vs_t *to )
{ int i;
	for(i=0;i<number_of_vs;i++)
		to[i]&= from[i];
}

int vs_test( const vs_t *test, const vs_t *vs )
{ int i;
	for(i=0;i<number_of_vs;i++)
		if( vs[i] & test[i] )
			return(1);
	return(0);
}

int vs_isclear( const vs_t *vs )
{ int i;
	for(i=0;i<number_of_vs;i++)
		if( vs[i] )
			return(0);
	return(1);
}

int vs_isfull( const vs_t *vs )
{ int i;
	for(i=0;i<number_of_vs;i++)
		if( vs[i]!=~((vs_t)0) )
			return(0);
	return(1);
}


